"""
Global Climate Assessment of Global Emission Scenarios.

Typically (but not always) produced by Integrated Assessment Models (IAMs).
"""

import importlib.metadata

__version__ = importlib.metadata.version("gcages")
